
from hmac import new as HMAC
import hashlib
from base64 import b64decode
from time import time as epoch

def hmac_from_name(alg_name='sha256'):
    assert alg_name in hashlib.algorithms, "Hash algorithm %s not known" % alg_name
    alg_constructor = getattr(hashlib, alg_name)
    return lambda key, msg=None: HMAC(key, msg, alg_constructor)

def _hmac(msg, alg_name, key):
    hm = hmac_from_name(alg_name)
    return hm(key, msg).hexdigest()

def hmac_wrap(msg, secret):
    assert not secret['accept_only'], "Key not valid for signing"
    handle, alg_name, key = [secret[k] for k in ('name', 'type', 'key')]
    timestamp = int(epoch())
    hm_digest = _hmac(handle +'\n' + str(timestamp) + '\n' + msg, alg_name, key)
    return {'msg': msg, 'hmac': hm_digest, 'handle': handle, 'timestamp': timestamp}

def lookup_from_secrets(secrets):
    lookup = {}
    for secret in secrets:
        secret = secret.copy()
        secret['key'] = b64decode(secret['key'])
        lookup[secret['name']] = secret
    return lookup

def signers_from_secrets(secrets):
    signers = []
    for secret in secrets:
        secret = secret.copy()
        secret['key'] = b64decode(secret['key'])
        if not secret['accept_only']:
            signers.append(secret)
    return signers

def assert_hmac(msg, lookup):
    assert 'hmac' in msg, "hmac field missing"
    assert 'msg' in msg, "msg field missing"
    assert 'handle' in msg, "handle field missing"
    assert 'timestamp' in msg, "timestamp field missing"
    handle = msg['handle']
    assert handle in lookup, "Handle unknown"
    alg_name, key, max_age_secs = [lookup[handle][k] for k in ('type', 'key', 'max_age_secs')]
    assert epoch() < msg['timestamp'] + max_age_secs, "Message expired"
    hm_digest = _hmac(handle + '\n' + str(msg['timestamp']) + '\n' + msg['msg'], alg_name, key)
    assert hm_digest == msg['hmac'], "Signature invalid"
    return True

import json

class AutoHMAC(object):
    """A fully-automatic HMAC sign-and-verify service
       NOTE: This class requires a CredClient object to operate, as provided by 
       credservice."""
    def __init__(self, tellme, resource='stomp-sig'):
        """Initialize an AutoHMAC instance from a CredClient object)"""
        secrets = tellme(resource)
        self.signers = signers_from_secrets(secrets)
        self.lookup = lookup_from_secrets(secrets)

    def pack(self, obj):
        """Pack an object for transmission"""
        raw_msg = json.dumps(obj)
        return hmac_wrap(raw_msg, self.signers[0])

    def unpack(self, msg):
        """Assert and unpack a signed message"""
        assert_hmac(msg, self.lookup)
        return json.loads(msg['msg'])




import logging, logging.handlers
from socket import error as socket_error

def add_argparse_group(parser):
    """Add a configuration group for plumb_util to an argparser"""
    group = parser.add_argument_group('find_service', 'SRV lookup configuration')
    group.add_argument('-D', '--domain', type = str, dest = 'zone', default = None,
                       help = 'DNS domain to consult for service autodiscovery.')
    group.add_argument('-L', '--loglevel', dest='log_level', default='error',
                       choices=['debug', 'info', 'warn', 'error'],
                       help = 'Set the syslog logging level.')

def init_logging(level, procname):
    logger = logging.root
    str_fmt = '%(asctime)s ' + procname + ': ' + logging.BASIC_FORMAT
    date_fmt = '%b %d %H:%M:%S'
    log_fmt = logging.Formatter(str_fmt, date_fmt)
    try:
        h1 = logging.handlers.SysLogHandler(address='/dev/log',
                facility=logging.handlers.SysLogHandler.LOG_LOCAL0)
    except socket_error:
        h1 = None

    h2 = logging.StreamHandler()
    h2.setFormatter(log_fmt)
    logger.addHandler(h2)
    if h1 is not None:
        h1.setFormatter(log_fmt)
        logger.addHandler(h1)
    logger.setLevel(level)
    if h1 is None:
        logger.warn('unable to initialize syslog logger')
    return logger

def logger_from_args(args, procname):
    # Set the log level
    if args.log_level == 'debug':
        level=logging.DEBUG
    elif args.log_level == 'info':
        level=logging.INFO
    elif args.log_level == 'warn':
        level=logging.WARN
    elif args.log_level == 'error':
        level=logging.ERROR
    return init_logging(level, procname)




from setuptools import setup, find_packages
import sys

setup(
    name="plumb-util",
    version = '0.1',
    maintainer='Luminoso, LLC',
    maintainer_email='dev@lumino.so',
    license = "MIT",
    url = 'http://github.com/LuminosoInsight/plumb_util',
    platforms = ["any"],
    description = "Plumbing utility library",
    packages=find_packages(),
    install_requires=['dnspython'],
)



import dns.resolver
import stomp
import random

def _weighted_shuffle(weight_item_pairs,prng=random):
    """Shuffle items from weight_item_pairs using weights"""
    pairs = [pair for pair in weight_item_pairs]
    pairs.sort(key=lambda pair: pair[0])
    total = sum(pair[0] for pair in pairs)

    choices = []
    for i in range(len(pairs)):
        # Detect a one-length array and take a shortcut
        if len(pairs) == 1:
            choices.append(pairs.pop(0)[1])
            return choices

        # Choose a score
        score = prng.randint(0, total-1)

        # Find the item the score selected
        for j in range(len(pairs)):
            score -= pairs[j][0]
            if score < 0:
                # Add the choice to the choices list
                choice = pairs.pop(j)
                choices.append(choice[1])

                # Update the PRNG ceiling
                total -= choice[0]
                break

    # This line should not be reached
    assert False, "The loop should have already returned."

def find_text(service, zone=None):
    """Return a TXT record for the specified service and zone"""
    service_name = service if zone is None else service + '.' + zone + '.'
    results = dns.resolver.query(service_name, 'TXT')
    text = None

    for record in results:
        text = record.strings[0]
        break

    return text

def find_service(service, zone=None):
    """Return a properly-weighted try list of servers for the specified zone"""
    service_name = service if zone is None else service + '.' + zone + '.'
    results = dns.resolver.query(service_name, 'SRV')

    # Group results by priority
    by_priority = {}
    for record in results:
        p_list = by_priority.setdefault(record.priority, [])
        p_list.append(record)

    # Weighted-random sort each priority tier
    choices = []
    for priority in sorted(by_priority.keys()):
        # Prepare the p_list for a weighted shuffle
        p_list = by_priority[priority]
        pairs = [(record.weight, (record.target.to_text(), record.port)) for record in p_list]

        # Add shuffled results to the choices list
        choices += _weighted_shuffle(pairs)

    return choices
